#include "Owned_string.h"

#include <algorithm>
#include <cstdlib>
#include <cstring>

Owned_string::Owned_string()
        : size_{0}
        , capacity_{0}
        , data_{nullptr}
{ }

Owned_string::Owned_string(const char* s)
    : Owned_string(s, s + strlen(s))
{ }

Owned_string::Owned_string(const char* begin, const char* end)
    : size_(begin <= end ? end - begin : 0)
    , capacity_{size_ ? size_ + 1 : 0}
    , data_{capacity_ ? new char[capacity_] : nullptr}
{
    if (data_) {
        std::copy(begin, end, data_);
        data_[size_] = '\0';
    }
}

Owned_string::Owned_string(Owned_string const& that)
        : Owned_string(that.begin(), that.end())
{ }

Owned_string::Owned_string(Owned_string&& that) noexcept
        : size_{that.size_}
        , capacity_{that.capacity_}
        , data_{that.data_}
{
    that.set_empty_();
}

/*
 * Destructor.
 */

Owned_string::~Owned_string()
{
    delete [] data_;
}

/*
 * Assignment “operators”.
 */

Owned_string& Owned_string::operator=(Owned_string const& that)
{
    if (this == &that) return *this;

    if (that.size_ > 0 && that.size_ + 1 >= capacity_) {
        char* new_data  = new char[that.size_ + 1];
        delete [] data_;
        data_     = new_data;
        capacity_ = that.size_ + 1;
    }

    if (data_) {
        if (that.data_)
            std::copy(that.begin(), that.end() + 1, data_);
        else
            data_[0] = '\0';
    }

    size_ = that.size_;

    return *this;
}

Owned_string& Owned_string::operator=(Owned_string&& that) noexcept
{
    Owned_string temp(std::move(that));
    swap(temp);
    return *this;
}

/*
 * Other `String` operations.
 */

bool Owned_string::empty() const
{
    return size_ == 0;
}

size_t Owned_string::size() const
{
    return size_;
}

char Owned_string::operator[](size_t index) const
{
    return data_[index];
}

char& Owned_string::operator[](size_t index)
{
    return data_[index];
}

void Owned_string::push_back(char c)
{
    // Make sure we have room for at least one more character:
    reserve(1);
    data_[size_++] = c;
    data_[size_] = '\0';
}

void Owned_string::pop_back()
{
    data_[--size_] = '\0';
}

void Owned_string::reserve(size_t additional_cap)
{
    if (size_ || additional_cap)
        ensure_capacity_(size_ + additional_cap + 1);
}

void Owned_string::swap(Owned_string& that)
{
    std::swap(size_, that.size_);
    std::swap(capacity_, that.capacity_);
    std::swap(data_, that.data_);
}

Owned_string& Owned_string::operator+=(Owned_string const& that)
{
    append_range_(that.begin(), that.end());
    return *this;
}

Owned_string& Owned_string::operator+=(const char* that)
{
    append_range_(that, that + strlen(that));
    return *this;
}

const char* Owned_string::c_str() const
{
    return data_? data_ : "";
}

Owned_string operator+(Owned_string const& a, Owned_string const& b)
{
    Owned_string result;
    result.reserve(a.size() + b.size());
    result += a;
    result += b;
    return result;
}

Owned_string operator+(Owned_string const& a, const char* b)
{
    Owned_string result;
    result.reserve(a.size() + std::strlen(b));
    result += a;
    result += b;
    return result;
}

Owned_string operator+(const char* a, Owned_string const& b)
{
    Owned_string result;
    result.reserve(std::strlen(a) + b.size());
    result += a;
    result += b;
    return result;
}

Owned_string operator+(Owned_string&& a, Owned_string const& b)
{
    a += b;
    return std::move(a);
}

Owned_string operator+(Owned_string&& a, const char* b)
{
    a += b;
    return std::move(a);
}

const char* Owned_string::begin() const
{
    return data_;
}

char* Owned_string::begin()
{
    return data_;
}

const char* Owned_string::end() const
{
    return data_ + size_;
}

char* Owned_string::end()
{
    return data_ + size_;
}

// Clears this `String` to be the empty string with no allocated memory.
// Note that if `data_` points to a live object then calling this
// function could leak it.
void Owned_string::set_empty_()
{
    size_ = capacity_ = 0;
    data_ = nullptr;
}

// Makes sure that this string has at least capacity `min_cap`, growing
// it if necessary.
void Owned_string::ensure_capacity_(size_t min_cap)
{
    if (capacity_ < min_cap) {
        size_t new_cap = std::max(min_cap, 2 * capacity_);
        char* new_data = new char[new_cap];
        if (data_)
            std::copy(begin(), end() + 1, new_data);
        delete [] data_;
        data_ = new_data;
        capacity_ = new_cap;
    }
}

void Owned_string::append_range_(const char* begin1, const char* end1)
{
    size_t size1 = end1 - begin1;
    reserve(size1 + 1);
    std::copy(begin1, end1, end());
    size_ += size1;
    *end() = '\0';
}
